@@ -34,7 +34,7 @@
  */
 #ifdef  HAVE_STDARG_H
 void
-logmsg(int priority, char *fmt, ...)
+logmsg(const int priority, const char *fmt, ...)
 {
     char    buf[MAXBUF + 1];
     va_list ap;
@@ -71,7 +71,7 @@ logmsg(int priority, char *fmt, ...)
 }
 #else
 void
-logmsg(int priority, char *fmt, va_alist)
+logmsg(const int priority, const char *fmt, va_alist)
 va_dcl
 {
     char    buf[MAXBUF + 1];
@@ -113,7 +113,7 @@ va_dcl
  * Translate inet address into a string
  */
 void
-addr2str(char *res, int res_len, struct in_addr *addr)
+addr2str(char *const res, const int res_len, const struct in_addr *addr)
 {
     char    *cp;
 
@@ -134,14 +134,14 @@ addr2str(char *res, int res_len, struct in_addr *addr)
  * Return a string representation for a back-end address
  */
 void
-str_be(char *buf, int max, BACKEND *be)
+str_be(char *const buf, const int max, const BACKEND *be)
 {
     char    tmp[MAXBUF];
 
     switch(be->domain) {
     case PF_INET:
         addr2str(tmp, MAXBUF - 1, &be->addr.in.sin_addr);
-        snprintf(buf, max, "%s:%hd", buf, ntohs(be->addr.in.sin_port));
+        snprintf(buf, max, "%s:%hd", tmp, ntohs(be->addr.in.sin_port));
         break;
     case PF_UNIX:
         strncpy(buf, be->addr.un.sun_path, max);
@@ -158,7 +158,7 @@ str_be(char *buf, int max, BACKEND *be)
  * return a code and possibly content in the arg
  */
 int
-check_header(char *header, char *content)
+check_header(const char *header, char *const content)
 {
     regmatch_t  matches[4];
     static struct {
@@ -174,6 +174,7 @@ check_header(char *header, char *content)
         { "Host",               4,  HEADER_HOST },
         { "Referer",            7,  HEADER_REFERER },
         { "User-agent",         10, HEADER_USER_AGENT },
+        { "Destination",        11, HEADER_DESTINATION },
         { "",                   0,  HEADER_OTHER },
     };
     int i;
@@ -199,7 +200,7 @@ check_header(char *header, char *content)
  * Find a session in a tree
  */
 static SESS *
-sess_find(SESS *root, char *key)
+sess_find(SESS *const root, const char *key)
 {
     int cmp;
 
@@ -216,7 +217,7 @@ sess_find(SESS *root, char *key)
  * Add a new session
  */
 static SESS *
-sess_add(SESS *root, char *key, BACKEND *to_host)
+sess_add(SESS *root, const char *key, BACKEND *const to_host)
 {
     int cmp;
 
@@ -278,7 +279,7 @@ sess_del(SESS *root)
  * Clean stale (expired) sessions
  */
 static SESS *
-sess_clean(SESS *root, time_t lim)
+sess_clean(SESS *root, const time_t lim)
 {
     if(root == NULL)
         return NULL;
@@ -294,7 +295,7 @@ sess_clean(SESS *root, time_t lim)
  * Clean dead back-ends
  */
 static SESS *
-sess_dead(SESS *root, BACKEND *be)
+sess_dead(SESS *root, const BACKEND *be)
 {
     if(root == NULL)
         return NULL;
@@ -344,7 +345,7 @@ sess_balance(SESS *root)
 }
 
 static int
-match_service(SERVICE *svc, char *request, char **headers)
+match_service(const SERVICE *svc, const char *request, char **const headers)
 {
     MATCHER *m;
     int     i, found;
@@ -379,18 +380,24 @@ match_service(SERVICE *svc, char *request, char **headers)
  * Find the right service for a request
  */
 SERVICE *
-get_service(LISTENER *lstn, char *request, char **headers)
+get_service(const LISTENER *lstn, const char *request, char **const headers)
 {
     SERVICE *svc;
 
-    for(svc = lstn->services; svc; svc = svc->next)
+    for(svc = lstn->services; svc; svc = svc->next) {
+        if(svc->disabled)
+            continue;
         if(match_service(svc, request, headers))
             return svc;
+    }
 
     /* try global services */
-    for(svc = services; svc; svc = svc->next)
+    for(svc = services; svc; svc = svc->next) {
+        if(svc->disabled)
+            continue;
         if(match_service(svc, request, headers))
             return svc;
+    }
 
     /* nothing matched */
     return NULL;
@@ -400,7 +407,7 @@ get_service(LISTENER *lstn, char *request, char **headers)
  * extract the session key for a given request
  */
 static int
-get_REQUEST(char *res, SERVICE *svc, char *request)
+get_REQUEST(char *res, const SERVICE *svc, const char *request)
 {
     int         n;
     regmatch_t  matches[4];
@@ -417,7 +424,7 @@ get_REQUEST(char *res, SERVICE *svc, char *request)
 }
 
 static int
-get_HEADERS(char *res, SERVICE *svc, char **headers)
+get_HEADERS(char *res, const SERVICE *svc, char **const headers)
 {
     int         i, n;
     regmatch_t  matches[4];
@@ -445,7 +452,7 @@ static BACKEND *
 rand_backend(BACKEND *be, int pri)
 {
     while(be) {
-        if(!be->alive) {
+        if(!be->alive || be->disabled) {
             be = be->next;
             continue;
         }
@@ -460,33 +467,34 @@ rand_backend(BACKEND *be, int pri)
  * Find the right back-end for a request
  */
 BACKEND *
-get_backend(SERVICE *svc, struct in_addr from_host, char *request, char **headers)
+get_backend(SERVICE *const svc, const struct in_addr *from_host, const char *request, char **const headers)
 {
     BACKEND     *res;
     SESS        *sp;
     char        key[KEY_SIZE + 1];
-    in_addr_t   addr;
-    int         pri;
+    int         ret_val;
 
-    /* blocked */
     if(svc->tot_pri <= 0)
-        return NULL;
+        /* it might be NULL, but that is OK */
+        return svc->emergency;
 
-    pthread_mutex_lock(&svc->mut);
+    if(ret_val = pthread_mutex_lock(&svc->mut))
+        logmsg(LOG_WARNING, "get_backend() lock: %s", strerror(ret_val));
     switch(svc->sess_type) {
     case SESS_NONE:
         /* choose one back-end randomly */
         res = rand_backend(svc->backends, random() % svc->tot_pri);
         break;
     case SESS_IP:
-        /* "sticky" mappings */
-        addr = from_host.s_addr;
-        pri = 0;
-        while(addr) {
-            pri = (pri << 3) ^ (addr & 0xff);
-            addr = (addr >> 8);
+        addr2str(key, KEY_SIZE, from_host);
+        if((sp = sess_find(svc->sessions, key)) == NULL) {
+            /* no session yet - create one */
+            res = rand_backend(svc->backends, random() % svc->tot_pri);
+            svc->sessions = sess_add(svc->sessions, key, res);
+        } else {
+            res = sp->to_host;
+            sp->last_acc = time(NULL);
         }
-        res = rand_backend(svc->backends, (pri & 0xffff) % svc->tot_pri);
         break;
     case SESS_PARM:
         if(get_REQUEST(key, svc, request)) {
@@ -518,7 +526,8 @@ get_backend(SERVICE *svc, struct in_addr from_host, char *request, char **header
         }
         break;
     }
-    pthread_mutex_unlock(&svc->mut);
+    if(ret_val = pthread_mutex_unlock(&svc->mut))
+        logmsg(LOG_WARNING, "get_backend() unlock: %s", strerror(ret_val));
 
     return res;
 }
@@ -527,17 +536,20 @@ get_backend(SERVICE *svc, struct in_addr from_host, char *request, char **header
  * (for cookies/header only) possibly create session based on response headers
  */
 void
-upd_session(SERVICE *svc, char **headers, BACKEND *be)
+upd_session(SERVICE *const svc, char **const headers, BACKEND *const be)
 {
     char            key[KEY_SIZE + 1];
+    int             ret_val;
 
     if(svc->sess_type != SESS_HEADER && svc->sess_type != SESS_COOKIE)
         return;
-    pthread_mutex_lock(&svc->mut);
+    if(ret_val = pthread_mutex_lock(&svc->mut))
+        logmsg(LOG_WARNING, "upd_session() lock: %s", strerror(ret_val));
     if(get_HEADERS(key, svc, headers))
         if(sess_find(svc->sessions, key) == NULL)
             svc->sessions = sess_add(svc->sessions, key, be);
-    pthread_mutex_unlock(&svc->mut);
+    if(ret_val = pthread_mutex_unlock(&svc->mut))
+        logmsg(LOG_WARNING, "upd_session() unlock: %s", strerror(ret_val));
     return;
 }
 
@@ -545,25 +557,83 @@ upd_session(SERVICE *svc, char **headers, BACKEND *be)
  * mark a backend host as dead; remove its sessions
  */
 void
-kill_be(SERVICE *svc, BACKEND *be)
+kill_be(SERVICE *const svc, const BACKEND *be)
 {
     BACKEND *b;
+    int     ret_val;
 
-    pthread_mutex_lock(&svc->mut);
+    if(ret_val = pthread_mutex_lock(&svc->mut))
+        logmsg(LOG_WARNING, "kill_be() lock: %s", strerror(ret_val));
     svc->tot_pri = 0;
     for(b = svc->backends; b; b = b->next) {
         if(b == be)
             b->alive = 0;
-        if(b->alive)
+        if(b->alive && !b->disabled)
             svc->tot_pri += b->priority;
     }
     svc->sessions = sess_dead(svc->sessions, be);
-    pthread_mutex_unlock(&svc->mut);
+    if(ret_val = pthread_mutex_unlock(&svc->mut))
+        logmsg(LOG_WARNING, "kill_be() unlock: %s", strerror(ret_val));
+    return;
+}
+
+/*
+ * Update the number of requests and time to answer for a given back-end
+ */
+void
+upd_be(BACKEND *const be, const double elapsed)
+{
+#ifndef NO_DYNSCALE
+    int     ret_val;
+
+    if(ret_val = pthread_mutex_lock(&be->mut))
+        logmsg(LOG_WARNING, "upd_be() lock: %s", strerror(ret_val));
+    be->t_requests += elapsed;
+    if(++be->n_requests > 2000) {
+        /* scale it down */
+        be->n_requests /= 2;
+        be->t_requests /= 2;
+    }
+    be->t_average = be->t_requests / be->n_requests;
+    if(ret_val = pthread_mutex_unlock(&be->mut))
+        logmsg(LOG_WARNING, "upd_be() unlock: %s", strerror(ret_val));
+#endif  /* NO_DYNSCALE */
+    return;
+}
+
+/*
+ * disable a backend; sessions are NOT affected
+ */
+static void
+disable_be(SERVICE *const svc, const BACKEND *be)
+{
+    BACKEND *b;
+    int     ret_val;
+
+    if(ret_val = pthread_mutex_lock(&svc->mut))
+        logmsg(LOG_WARNING, "disable_be() lock: %s", strerror(ret_val));
+    svc->tot_pri = 0;
+    for(b = svc->backends; b; b = b->next) {
+        if(b == be)
+            b->disabled = 1;
+        if(b->alive && !b->disabled)
+            svc->tot_pri += b->priority;
+    }
+    if(ret_val = pthread_mutex_unlock(&svc->mut))
+        logmsg(LOG_WARNING, "disable_be() unlock: %s", strerror(ret_val));
     return;
 }
 
 static pthread_mutex_t  host_mut;       /* mutex to protect gethostbyname */
 
+void
+init_host_mut(void)
+{
+    /* pthread_mutex_init() always returns 0 */
+    pthread_mutex_init(&host_mut, NULL);
+    return;
+}
+
 /*
  * Find if a redirect needs rewriting
  * In general we have two possibilities that require it:
@@ -571,12 +641,17 @@ static pthread_mutex_t  host_mut;       /* mutex to protect gethostbyname */
  * (2) if the redirect was done to the back-end rather than the listener
  */
 int
-need_rewrite(char *location, char *path, LISTENER *lstn, BACKEND *be)
+need_rewrite(const int rewr_loc, char *const location, char *const path, const LISTENER *lstn, const BACKEND *be)
 {
     struct sockaddr_in  addr;
     struct hostent      *he;
     regmatch_t          matches[4];
     char                *proto, *host, *port;
+    int                 ret_val;
+
+    /* check if rewriting is required at all */
+    if(rewr_loc == 0)
+        return 0;
 
     /* applies only to INET back-ends */
     if(be->domain != PF_INET)
@@ -587,6 +662,8 @@ need_rewrite(char *location, char *path, LISTENER *lstn, BACKEND *be)
         return 0;
     proto = location + matches[1].rm_so;
     host = location + matches[2].rm_so;
+    if(location[matches[3].rm_so] == '/')
+        matches[3].rm_so++;
     strcpy(path, location + matches[3].rm_so);
     location[matches[1].rm_eo] = location[matches[2].rm_eo] = '\0';
     if((port = strchr(host, ':')) != NULL)
@@ -598,16 +675,20 @@ need_rewrite(char *location, char *path, LISTENER *lstn, BACKEND *be)
     memset(&addr, 0, sizeof(addr));
     addr.sin_family = AF_INET;
     /* this is to avoid the need for gethostbyname_r */
-    pthread_mutex_lock(&host_mut);
+    if(ret_val = pthread_mutex_lock(&host_mut))
+        logmsg(LOG_WARNING, "need_rewrite() lock: %s", strerror(ret_val));
     if((he = gethostbyname(host)) == NULL || he->h_addr_list[0] == NULL) {
-        pthread_mutex_unlock(&host_mut);
+        logmsg(LOG_WARNING, "gethostbyname(%s): %s", host, hstrerror(h_errno));
+        if(ret_val = pthread_mutex_unlock(&host_mut))
+            logmsg(LOG_WARNING, "need_rewrite() unlock: %s", strerror(ret_val));
         return 0;
     }
     /*
      * prepare the address
      */
     memcpy(&addr.sin_addr.s_addr, he->h_addr_list[0], sizeof(addr.sin_addr.s_addr));
-    pthread_mutex_unlock(&host_mut);
+    if(ret_val = pthread_mutex_unlock(&host_mut))
+        logmsg(LOG_WARNING, "need_rewrite() unlock: %s", strerror(ret_val));
 
     if(port)
         addr.sin_port = (in_port_t)htons(atoi(port));
@@ -621,12 +702,14 @@ need_rewrite(char *location, char *path, LISTENER *lstn, BACKEND *be)
     if(memcmp(&be->addr.in.sin_addr.s_addr, &addr.sin_addr.s_addr, sizeof(addr.sin_addr.s_addr)) == 0
     && memcmp(&be->addr.in.sin_port, &addr.sin_port, sizeof(addr.sin_port)) == 0)
         return 1;
-    /*
-     * check if the Location points to the Listener but with the wrong port
-     */
-    if(memcmp(&lstn->addr.sin_addr.s_addr, &addr.sin_addr.s_addr, sizeof(addr.sin_addr.s_addr)) == 0
-    && memcmp(&lstn->addr.sin_port, &addr.sin_port, sizeof(addr.sin_port)) != 0)
-        return 1;
+    if(rewr_loc == 1) {
+        /*
+         * check if the Location points to the Listener but with the wrong port
+         */
+        if(memcmp(&lstn->addr.sin_addr.s_addr, &addr.sin_addr.s_addr, sizeof(addr.sin_addr.s_addr)) == 0
+        && memcmp(&lstn->addr.sin_port, &addr.sin_port, sizeof(addr.sin_port)) != 0)
+            return 1;
+    }
     return 0;
 }
 
@@ -635,18 +718,18 @@ need_rewrite(char *location, char *path, LISTENER *lstn, BACKEND *be)
  * it will time-out after a much shorter time period SERVER_TO
  */
 int
-connect_nb(int sockfd, struct sockaddr *serv_addr, socklen_t addrlen, int to)
+connect_nb(const int sockfd, const struct sockaddr *serv_addr, const socklen_t addrlen, const int to)
 {
     int             flags, res, error;
     socklen_t       len;
     struct pollfd   p;
 
     if((flags = fcntl(sockfd, F_GETFL, 0)) < 0) {
-        logmsg(LOG_ERR, "fcntl GETFL failed: %s", strerror(errno));
+        logmsg(LOG_WARNING, "fcntl GETFL failed: %s", strerror(errno));
         return -1;
     }
     if(fcntl(sockfd, F_SETFL, flags | O_NONBLOCK) < 0) {
-        logmsg(LOG_ERR, "fcntl SETFL failed: %s", strerror(errno));
+        logmsg(LOG_WARNING, "fcntl SETFL failed: %s", strerror(errno));
         return -1;
     }
 
@@ -658,7 +741,7 @@ connect_nb(int sockfd, struct sockaddr *serv_addr, socklen_t addrlen, int to)
     if(res == 0) {
         /* connect completed immediately (usually localhost) */
         if(fcntl(sockfd, F_SETFL, flags) < 0) {
-            logmsg(LOG_ERR, "fcntl reSETFL failed: %s", strerror(errno));
+            logmsg(LOG_WARNING, "fcntl reSETFL failed: %s", strerror(errno));
             return -1;
         }
         return 0;
@@ -678,13 +761,13 @@ connect_nb(int sockfd, struct sockaddr *serv_addr, socklen_t addrlen, int to)
     /* socket is writeable == operation completed */
     len = sizeof(error);
     if(getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
-        logmsg(LOG_ERR, "getsockopt failed: %s", strerror(errno));
+        logmsg(LOG_WARNING, "getsockopt failed: %s", strerror(errno));
         return -1;
     }
 
     /* restore file status flags */
     if(fcntl(sockfd, F_SETFL, flags) < 0) {
-        logmsg(LOG_ERR, "fcntl reSETFL failed: %s", strerror(errno));
+        logmsg(LOG_WARNING, "fcntl reSETFL failed: %s", strerror(errno));
         return -1;
     }
 
@@ -713,6 +796,8 @@ thr_resurect(void *arg)
     time_t      last_time, cur_time;
     int         n, sock;
     char        buf[MAXBUF];
+    int         ret_val;
+    double      average, sq_average;
 
     for(last_time = time(NULL) - alive_to;;) {
         cur_time = time(NULL);
@@ -724,17 +809,21 @@ thr_resurect(void *arg)
         for(lstn = listeners; lstn; lstn = lstn->next)
         for(svc = lstn->services; svc; svc = svc->next)
             if(svc->sess_type != SESS_NONE) {
-                pthread_mutex_lock(&svc->mut);
+                if(ret_val = pthread_mutex_lock(&svc->mut))
+                    logmsg(LOG_WARNING, "thr_resurect() lock: %s", strerror(ret_val));
                 svc->sessions = sess_clean(svc->sessions, last_time - svc->sess_ttl);
                 svc->sessions = sess_balance(svc->sessions);
-                pthread_mutex_unlock(&svc->mut);
+                if(ret_val = pthread_mutex_unlock(&svc->mut))
+                    logmsg(LOG_WARNING, "thr_resurect() unlock: %s", strerror(ret_val));
             }
         for(svc = services; svc; svc = svc->next)
             if(svc->sess_type != SESS_NONE) {
-                pthread_mutex_lock(&svc->mut);
+                if(ret_val = pthread_mutex_lock(&svc->mut))
+                    logmsg(LOG_WARNING, "thr_resurect() lock: %s", strerror(ret_val));
                 svc->sessions = sess_clean(svc->sessions, last_time - svc->sess_ttl);
                 svc->sessions = sess_balance(svc->sessions);
-                pthread_mutex_unlock(&svc->mut);
+                if(ret_val = pthread_mutex_unlock(&svc->mut))
+                    logmsg(LOG_WARNING, "thr_resurect() unlock: %s", strerror(ret_val));
             }
 
         /* check hosts still alive - HAport */
@@ -756,7 +845,7 @@ thr_resurect(void *arg)
             if(connect_nb(sock, (struct sockaddr *)&be->HA, (socklen_t)sizeof(be->HA), be->to) != 0) {
                 kill_be(svc, be);
                 addr2str(buf, MAXBUF - 1, &be->HA.sin_addr);
-                logmsg(LOG_ERR,"BackEnd %s:%hd is dead (HA)", buf, ntohs(be->HA.sin_port));
+                logmsg(LOG_NOTICE, "BackEnd %s:%hd is dead (HA)", buf, ntohs(be->HA.sin_port));
             }
             shutdown(sock, 2);
             close(sock);
@@ -777,7 +866,7 @@ thr_resurect(void *arg)
             if(connect_nb(sock, (struct sockaddr *)&be->HA, (socklen_t)sizeof(be->HA), be->to) != 0) {
                 kill_be(svc, be);
                 addr2str(buf, MAXBUF - 1, &be->HA.sin_addr);
-                logmsg(LOG_ERR,"BackEnd %s:%hd is dead (HA)", buf, ntohs(be->HA.sin_port));
+                logmsg(LOG_NOTICE, "BackEnd %s:%hd is dead (HA)", buf, ntohs(be->HA.sin_port));
             }
             shutdown(sock, 2);
             close(sock);
@@ -803,19 +892,21 @@ thr_resurect(void *arg)
                     be->alive = 1;
                     if(be->domain == PF_INET) {
                         addr2str(buf, MAXBUF - 1, &be->addr.in.sin_addr);
-                        logmsg(LOG_ERR,"BackEnd %s:%hd resurrect", buf, ntohs(be->addr.in.sin_port));
+                        logmsg(LOG_NOTICE, "BackEnd %s:%hd resurrect", buf, ntohs(be->addr.in.sin_port));
                     } else
-                        logmsg(LOG_ERR,"BackEnd %s resurrect", be->addr.un.sun_path);
+                        logmsg(LOG_NOTICE, "BackEnd %s resurrect", be->addr.un.sun_path);
                 }
                 shutdown(sock, 2);
                 close(sock);
             }
-            pthread_mutex_lock(&svc->mut);
+            if(ret_val = pthread_mutex_lock(&svc->mut))
+                logmsg(LOG_WARNING, "thr_resurect() lock: %s", strerror(ret_val));
             svc->tot_pri = 0;
             for(be = svc->backends; be; be = be->next)
-                if(be->alive)
+                if(be->alive && !be->disabled)
                     svc->tot_pri += be->priority;
-            pthread_mutex_unlock(&svc->mut);
+            if(ret_val = pthread_mutex_unlock(&svc->mut))
+                logmsg(LOG_WARNING, "thr_resurect() unlock: %s", strerror(ret_val));
         }
         for(svc = services; svc; svc = svc->next) {
             for(be = svc->backends; be; be = be->next) {
@@ -835,18 +926,102 @@ thr_resurect(void *arg)
                 if(connect_nb(sock, addr, (socklen_t)sizeof(*addr), be->to) == 0) {
                     be->alive = 1;
                     str_be(buf, MAXBUF - 1, be);
-                    logmsg(LOG_ERR,"BackEnd %s resurrect", buf);
+                    logmsg(LOG_NOTICE, "BackEnd %s resurrect", buf);
                 }
                 shutdown(sock, 2);
                 close(sock);
             }
-            pthread_mutex_lock(&svc->mut);
+            if(ret_val = pthread_mutex_lock(&svc->mut))
+                logmsg(LOG_WARNING, "thr_resurect() lock: %s", strerror(ret_val));
             svc->tot_pri = 0;
             for(be = svc->backends; be; be = be->next)
-                if(be->alive)
+                if(be->alive && !be->disabled)
                     svc->tot_pri += be->priority;
-            pthread_mutex_unlock(&svc->mut);
+            if(ret_val = pthread_mutex_unlock(&svc->mut))
+                logmsg(LOG_WARNING, "thr_resurect() unlock: %s", strerror(ret_val));
         }
+
+#ifndef NO_DYNSCALE
+        /* scale the back-end priorities */
+        for(lstn = listeners; lstn; lstn = lstn->next)
+        for(svc = lstn->services; svc; svc = svc->next) {
+            average = sq_average = 0.0;
+            n = 0;
+            for(be = svc->backends; be; be = be->next) {
+                if(be->be_type != BACK_END || !be->alive || be->disabled)
+                    continue;
+                average += be->t_average;
+                sq_average += be->t_average * be->t_average;
+                n++;
+            }
+            if(n <= 1)
+                continue;
+            sq_average /= n;
+            average /= n;
+            sq_average = sqrt(sq_average - average * average);  /* this is now the standard deviation */
+            if(ret_val = pthread_mutex_lock(&svc->mut)) {
+                logmsg(LOG_WARNING, "thr_resurect() lock: %s", strerror(ret_val));
+                continue;
+            }
+            for(be = svc->backends; be; be = be->next) {
+                if(be->be_type != BACK_END || !be->alive || be->disabled)
+                    continue;
+                if(be->t_average < (average - sq_average)) {
+                    be->priority++;
+                    svc->tot_pri++;
+                    be->n_requests /= 4;
+                    be->t_requests /= 4;
+                }
+                if(be->t_average > (average + sq_average) && be->priority > 1) {
+                    be->priority--;
+                    svc->tot_pri--;
+                    be->n_requests /= 4;
+                    be->t_requests /= 4;
+                }
+            }
+            if(ret_val = pthread_mutex_unlock(&svc->mut))
+                logmsg(LOG_WARNING, "thr_resurect() unlock: %s", strerror(ret_val));
+        }
+
+        for(svc = services; svc; svc = svc->next) {
+            average = sq_average = 0.0;
+            n = 0;
+            for(be = svc->backends; be; be = be->next) {
+                if(be->be_type != BACK_END || !be->alive || be->disabled)
+                    continue;
+                average += be->t_average;
+                sq_average += be->t_average * be->t_average;
+                n++;
+            }
+            if(n <= 1)
+                continue;
+            sq_average /= n;
+            average /= n;
+            sq_average = sqrt(sq_average - average * average);  /* this is now the standard deviation */
+            if(ret_val = pthread_mutex_lock(&svc->mut)) {
+                logmsg(LOG_WARNING, "thr_resurect() lock: %s", strerror(ret_val));
+                continue;
+            }
+            for(be = svc->backends; be; be = be->next) {
+                if(be->be_type != BACK_END || !be->alive || be->disabled)
+                    continue;
+                if(be->t_average < (average - sq_average)) {
+                    be->priority++;
+                    svc->tot_pri++;
+                    be->n_requests /= 4;
+                    be->t_requests /= 4;
+                }
+                if(be->t_average > (average + sq_average) && be->priority > 1) {
+                    be->priority--;
+                    svc->tot_pri--;
+                    be->n_requests /= 4;
+                    be->t_requests /= 4;
+                }
+            }
+            if(ret_val = pthread_mutex_unlock(&svc->mut))
+                logmsg(LOG_WARNING, "thr_resurect() unlock: %s", strerror(ret_val));
+        }
+#endif  /* NO_DYNSCALE */
     }
 }
 
@@ -858,13 +1033,16 @@ static RSA              *RSA1024_keys[N_RSA_KEYS];  /* ephemeral RSA keys */
  * return a pre-generated RSA key
  */
 RSA *
-RSA_tmp_callback(SSL *ssl, int is_export, int keylength)
+RSA_tmp_callback(/* not used */SSL *ssl, /* not used */int is_export, int keylength)
 {
     RSA *res;
+    int ret_val;
 
-    pthread_mutex_lock(&RSA_mut);
+    if(ret_val = pthread_mutex_lock(&RSA_mut))
+        logmsg(LOG_WARNING, "RSA_tmp_callback() lock: %s", strerror(ret_val));
     res = (keylength <= 512)? RSA512_keys[rand() % N_RSA_KEYS]: RSA1024_keys[rand() % N_RSA_KEYS];
-    pthread_mutex_unlock(&RSA_mut);
+    if(ret_val = pthread_mutex_unlock(&RSA_mut))
+        logmsg(LOG_WARNING, "RSA_tmp_callback() unlock: %s", strerror(ret_val));
     return res;
 }
 
@@ -878,14 +1056,15 @@ init_RSAgen()
 
     for(n = 0; n < N_RSA_KEYS; n++) {
         if((RSA512_keys[n] = RSA_generate_key(512, RSA_F4, NULL, NULL)) == NULL) {
-            logmsg(LOG_ERR,"RSA_generate(%d, 512) failed", n);
+            logmsg(LOG_WARNING,"RSA_generate(%d, 512) failed", n);
             return;
         }
         if((RSA1024_keys[n] = RSA_generate_key(1024, RSA_F4, NULL, NULL)) == NULL) {
-            logmsg(LOG_ERR,"RSA_generate(%d, 1024) failed", n);
+            logmsg(LOG_WARNING,"RSA_generate(%d, 1024) failed", n);
             return;
         }
     }
+    /* pthread_mutex_init() always returns 0 */
     pthread_mutex_init(&RSA_mut, NULL);
     return;
 }
@@ -897,17 +1076,217 @@ init_RSAgen()
 void *
 thr_RSAgen(void *arg)
 {
-    int n;
+    int n, ret_val;
 
     for(;;) {
         sleep(T_RSA_KEYS);
-        pthread_mutex_lock(&RSA_mut);
+        if(ret_val = pthread_mutex_lock(&RSA_mut))
+            logmsg(LOG_WARNING, "thr_RSAgen() lock: %s", strerror(ret_val));
         for(n = 0; n < N_RSA_KEYS; n++) {
             RSA_free(RSA512_keys[n]);
             RSA512_keys[n] = RSA_generate_key(512, RSA_F4, NULL, NULL);
             RSA_free(RSA1024_keys[n]);
             RSA1024_keys[n] = RSA_generate_key(1024, RSA_F4, NULL, NULL);
         }
-        pthread_mutex_unlock(&RSA_mut);
+        if(ret_val = pthread_mutex_unlock(&RSA_mut))
+            logmsg(LOG_WARNING, "thr_RSAgen() unlock: %s", strerror(ret_val));
+    }
+}
+
+/*
+ * write sessions to the control socket
+ */
+static void
+dump_sess(const int control_sock, const SESS *sess, BACKEND *const backends)
+{
+    SESS    s;
+    BACKEND *be;
+    int     n_be;
+
+    if(sess) {
+        dump_sess(control_sock, sess->left, backends);
+        s = *sess;
+        for(n_be = 0, be = backends; be; be = be->next, n_be++)
+            if(be == s.to_host)
+                break;
+        if(!be)
+            /* should NEVER happen */
+            n_be = 0;
+        s.to_host = (BACKEND *)n_be;
+        write(control_sock, (void *)&s, sizeof(SESS));
+        dump_sess(control_sock, sess->right, backends);
+    }
+    return;
+}
+
+/*
+ * given a command, select a listener
+ */
+static LISTENER *
+sel_lstn(const CTRL_CMD *cmd)
+{
+    LISTENER    *lstn;
+    int         i;
+
+    if(cmd->listener < 0)
+        return NULL;
+    for(i = 0, lstn = listeners; lstn && i < cmd->listener; i++, lstn = lstn->next)
+        ;
+    return lstn;
+}
+
+/*
+ * given a command, select a service
+ */
+static SERVICE *
+sel_svc(const CTRL_CMD *cmd)
+{
+    SERVICE     *svc;
+    LISTENER    *lstn;
+    int         i;
+
+    if(cmd->listener < 0) {
+        svc = services;
+    } else {
+        if((lstn = sel_lstn(cmd)) == NULL)
+            return NULL;
+        svc = lstn->services;
+    }
+    for(i = 0; svc && i < cmd->service; i++, svc = svc->next)
+        ;
+    return svc;
+}
+
+/*
+ * given a command, select a back-end
+ */
+static BACKEND *
+sel_be(const CTRL_CMD *cmd)
+{
+    BACKEND     *be;
+    SERVICE     *svc;
+    int         i;
+
+    if((svc = sel_svc(cmd)) == NULL)
+        return NULL;
+    for(i = 0, be = svc->backends; be && i < cmd->backend; i++, be = be->next)
+        ;
+    return be;
+}
+
+/*
+ * The controlling thread
+ * listens to client requests and calls the appropriate functions
+ */
+void *
+thr_control(void *arg)
+{
+    CTRL_CMD        cmd;
+    struct sockaddr sa;
+    int             ctl, dummy;
+    LISTENER        *lstn, dummy_lstn;
+    SERVICE         *svc, dummy_svc;
+    BACKEND         *be, dummy_be;
+    SESS            dummy_sess;
+
+    /* just to be safe */
+    if(control_sock < 0)
+        return NULL;
+    memset(&dummy_lstn, 0, sizeof(dummy_lstn));
+    dummy_lstn.disabled = -1;
+    memset(&dummy_svc, 0, sizeof(dummy_svc));
+    dummy_svc.disabled = -1;
+    memset(&dummy_be, 0, sizeof(dummy_be));
+    dummy_be.disabled = -1;
+    memset(&dummy_sess, 0, sizeof(dummy_sess));
+    dummy_sess.to_host = (BACKEND *)-1;
+    for(;;) {
+        if((ctl = accept(control_sock, &sa, (socklen_t *)&dummy)) < 0) {
+            logmsg(LOG_WARNING, "thr_control() accept: %s", strerror(errno));
+            continue;
+        }
+        if(read(ctl, &cmd, sizeof(cmd)) != sizeof(cmd)) {
+            logmsg(LOG_WARNING, "thr_control() read: %s", strerror(errno));
+            continue;
+        }
+        switch(cmd.cmd) {
+        case CTRL_LST:
+            /* logmsg(LOG_INFO, "thr_control() list"); */
+            for(lstn = listeners; lstn; lstn = lstn->next) {
+                write(ctl, (void *)lstn, sizeof(LISTENER));
+                for(svc = lstn->services; svc; svc = svc->next) {
+                    write(ctl, (void *)svc, sizeof(SERVICE));
+                    for(be = svc->backends; be; be = be->next)
+                        write(ctl, (void *)be, sizeof(BACKEND));
+                    write(ctl, (void *)&dummy_be, sizeof(BACKEND));
+                    if(dummy = pthread_mutex_lock(&svc->mut))
+                        logmsg(LOG_WARNING, "thr_control() lock: %s", strerror(dummy));
+                    else {
+                        dump_sess(ctl, svc->sessions, svc->backends);
+                        if(dummy = pthread_mutex_unlock(&svc->mut))
+                            logmsg(LOG_WARNING, "thr_control() unlock: %s", strerror(dummy));
+                    }
+                    write(ctl, (void *)&dummy_sess, sizeof(SESS));
+                }
+                write(ctl, (void *)&dummy_svc, sizeof(SERVICE));
+            }
+            write(ctl, (void *)&dummy_lstn, sizeof(LISTENER));
+            for(svc = services; svc; svc = svc->next) {
+                write(ctl, (void *)svc, sizeof(SERVICE));
+                for(be = svc->backends; be; be = be->next)
+                    write(ctl, (void *)be, sizeof(BACKEND));
+                write(ctl, (void *)&dummy_be, sizeof(BACKEND));
+                if(dummy = pthread_mutex_lock(&svc->mut))
+                    logmsg(LOG_WARNING, "thr_control() lock: %s", strerror(dummy));
+                else {
+                    dump_sess(ctl, svc->sessions, svc->backends);
+                    if(dummy = pthread_mutex_unlock(&svc->mut))
+                        logmsg(LOG_WARNING, "thr_control() unlock: %s", strerror(dummy));
+                }
+                write(ctl, (void *)&dummy_sess, sizeof(SESS));
+            }
+            write(ctl, (void *)&dummy_svc, sizeof(SERVICE));
+            break;
+        case CTRL_EN_LSTN:
+            if((lstn = sel_lstn(&cmd)) == NULL)
+                logmsg(LOG_INFO, "thr_control() bad listener %d", cmd.listener);
+            else
+                lstn->disabled = 0;
+            break;
+        case CTRL_DE_LSTN:
+            if((lstn = sel_lstn(&cmd)) == NULL)
+                logmsg(LOG_INFO, "thr_control() bad listener %d", cmd.listener);
+            else
+                lstn->disabled = 1;
+            break;
+        case CTRL_EN_SVC:
+            if((svc = sel_svc(&cmd)) == NULL)
+                logmsg(LOG_INFO, "thr_control() bad service %d/%d", cmd.listener, cmd.service);
+            else
+                svc->disabled = 0;
+            break;
+        case CTRL_DE_SVC:
+            if((svc = sel_svc(&cmd)) == NULL)
+                logmsg(LOG_INFO, "thr_control() bad service %d/%d", cmd.listener, cmd.service);
+            else
+                svc->disabled = 1;
+            break;
+        case CTRL_EN_BE:
+            if((be = sel_be(&cmd)) == NULL)
+                logmsg(LOG_INFO, "thr_control() bad backend %d/%d/%d", cmd.listener, cmd.service, cmd.backend);
+            else
+                be->disabled = 0;
+            break;
+        case CTRL_DE_BE:
+            if((be = sel_be(&cmd)) == NULL)
+                logmsg(LOG_INFO, "thr_control() bad backend %d/%d/%d", cmd.listener, cmd.service, cmd.backend);
+            else
+                be->disabled = 1;
+            break;
+        default:
+            logmsg(LOG_WARNING, "thr_control() unknown command");
+            break;
+        }
+        close(ctl);
     }
 }
